"""
    Name:           onboardme.misc
    DESCRIPTION:    this is where I'm putting extra functions
    AUTHOR:         https://github.com/jessebot
    LICENSE:        GNU AFFERO GENERAL PUBLIC LICENSE Version 3
"""
from .subproc import subproc
from .console_logging import print_header


def map_caps_to_control():
    """
    Maps capslock to control. This is ugly and awful and untested
    """
    print_header("⌨️  Mapping capslock to control...")
    subproc(["setxkbmap -layout us -option ctrl:nocaps"])


def setup_cronjobs():
    """
    setup any important cronjobs/alarms.
    Currently just adds nightly updates and reminders to take breaks
    """
    print_header("⏰ Installing new cronjobs...")
    print("\n")
    # TODO: is there a python cron library 🤔 install .cron dir? (is there a
    # standard in where cronjobs live for users [preferably in the home dir?)
    # that works on both macos and debian?



#!/usr/bin/env python3.11
"""
       Name:
DESCRIPTION:
     AUTHOR:
    LICENSE: GNU AFFERO GENERAL PUBLIC LICENSE Version 3
"""
from importlib.metadata import version as get_version
import logging as log
from xdg_base_dirs import xdg_config_home
from os import getenv, path, uname
from pathlib import Path
import wget
import yaml


# version of onboardme
VERSION = get_version('onboardme')

# pathing
XDG_CONFIG_DIR = xdg_config_home()
PWD = path.dirname(__file__)
HOME_DIR = getenv("HOME")

# env
SYSINFO = uname()
# this will be something like ('Darwin', 'x86_64')
OS = (SYSINFO.sysname, SYSINFO.machine)

# step config is different per OS
STEPS = ['dot_files','packages','font_setup','neovim_setup','group_setup']
if OS[0] == 'Darwin':
    STEPS.append('sudo_setup')

# package manager config is different per OS
PKG_MNGRS = ['brew','pip3.11']
if OS[0] == 'Linux':
    PKG_MNGRS.extend(['apt','snap','flatpak'])


default_dotfiles = ("https://raw.githubusercontent.com/jessebot/dot_files/"
                    "main/.config/onboardme/config.yml")


def load_cfg(config_file='config.yml') -> dict:
    """
    load yaml config files for onboardme
    """
    config_dir = path.join(xdg_config_home(), 'onboardme')
    config_full_path = path.join(config_dir, config_file)

    # create default pathing and config file if it doesn't exist
    if not path.exists(config_full_path):
        Path(config_dir).mkdir(parents=True, exist_ok=True)
        # downloads a default config file from default dot files
        wget.download(default_dotfiles, config_full_path)

    with open(config_full_path, 'r') as yaml_file:
        return yaml.safe_load(yaml_file)


USR_CONFIG_FILE = load_cfg()



"""
config variable processing library for onboardme
"""

import logging as log

# rich helps pretty print everything
from rich.prompt import Confirm

# custom libs
from .constants import OS, USR_CONFIG_FILE
from .console_logging import print_panel


def check_os_support():
    """
    verify we're on a supported OS and ask to quit if not.
    """
    if OS[0] not in ('Linux', 'Darwin'):
        msg = (f"[ohno]{OS}[/ohno] isn't officially supported. We"
               " have only tested Debian, Ubuntu, and macOS.")
        print_panel(msg, "⚠️  [warn]WARNING")

        quit_y = Confirm.ask("🌊 You're in uncharted waters. Wanna quit?")
        if quit_y:
            print_panel("That's probably safer. Have a safe day, friend.",
                        "Safety Award ☆ ")
            quit()
        else:
            print_panel("[red]Yeehaw, I guess.", "¯\\_(ツ)_/¯")
    else:
        print_panel("Operating System and Architechure [green]supported ♥",
                    "[cornflower_blue]Compatibility Check")


def process_steps(steps: list, firewall=False, browser=False) -> list:
    """
    process which steps to run for which OS, which steps the user passed in,
    and then make sure dependent steps are always run.

    Returns a list of str type steps to run.

    # TODO: if 'capslock_to_control' in steps: map_caps_to_control()
    """
    if steps:
        steps_list = list(steps)
        # setting up neovim is useless if we don't have an init.vim or init.lua
        if 'neovim_setup' in steps_list and 'dot_files' not in steps_list:
            steps_list.append('dot_files')

        if firewall and 'Linux' in OS:
            # currently don't have a great firewall on macOS, sans lulu
            steps_list.append('firewall_setup')

    removed_duplicates = set(steps_list)
    steps = list(removed_duplicates)
    default_order = ['dot_files', 'packages', 'font_setup', 'neovim_setup']
    if OS[0] == 'Linux':
        default_order.append('group_setup')
    else:
        default_order.append('sudo_setup')

    # Rearrange list by other list order Using list comprehension
    return [ele for ele in default_order if ele in steps]


def sort_pkgmngrs(package_managers_list: list) -> list:
    """
    make sure the package managers are in the right order 🤦
    e.g. apt installs snap and flatpak, so niether can be run until apt is run

    Takes list of package manager str and reorders them be (if they exist):
       ['brew', 'pip3.11', 'apt', 'snap', 'flatpak']
    """
    pkg_mngr_default_order = ['brew', 'pip3.11', 'apt', 'snap', 'flatpak']

    # Rearrange list by other list order Using list comprehension
    return [ele for ele in pkg_mngr_default_order if ele in package_managers_list]


def fill_in_defaults(defaults: dict, user_config: dict,
                     always_prefer_default=False) -> dict:
    """
    Compares/Combines a default dict and another dict. Prefer default values
    only if the value is empty in the second dict. Then return new dict.
    """
    for key, value in user_config.items():
        # we have to iterate through the entire config file, and who knows how
        # many levels there are, so we use recursion of this function
        if type(value) is dict:
            result_config = fill_in_defaults(defaults[key], user_config[key],
                                             always_prefer_default)
            user_config[key] = result_config

        if not value or always_prefer_default:
            user_config[key] = defaults[key]

    return user_config


def process_configs(overwrite: bool, repo: str, git_branch: str, 
                    pkg_mngrs: list, pkg_groups: list, firewall: bool,
                    remote_host: str, steps: list, log_file: str,
                    log_level: str) -> dict:
    """
    process the config in ~/.config/onboardme/config.yml if it exists,
    then process the cli dict, and fill in defaults for anything not explicitly
    defined. Returns full final config as dict for use in script.
    """
    if remote_host:
        firewall = True
        if type(remote_host) is str:
            remote_host = [remote_host]


    cli_dict = {'package': {'managers': {OS[0]: pkg_mngrs},
                            'groups': pkg_groups},
                'log': {'file': log_file, 'level': log_level},
                'remote_hosts': remote_host,
                'firewall': firewall,
                'steps': {OS[0]: steps},
                'dot_files': {'overwrite': overwrite,
                              'git_url': repo, 'git_branch': git_branch}}

    log.debug(f"cli_dict is:\n{cli_dict}\n")

    if OS[0] == 'Darwin':
        USR_CONFIG_FILE['package']['managers'].pop('Linux')
        USR_CONFIG_FILE['steps'].pop('Linux')
    else:
        USR_CONFIG_FILE['package']['managers'].pop('Darwin')
        USR_CONFIG_FILE['steps'].pop('Darwin')

    log.debug(f"🗂 ⚙️  user_config_file: \n{USR_CONFIG_FILE}\n")
    final_defaults = fill_in_defaults(cli_dict, USR_CONFIG_FILE, True)

    log.debug(" final config after filling cli_dict in with defaults:\n"
              f"{final_defaults}\n")

    # make sure the steps are in a valid order
    valid_steps = process_steps(final_defaults['steps'][OS[0]],
                                final_defaults['remote_hosts'])
    final_defaults['steps'][OS[0]] = valid_steps

    # make sure the package managers are in a valid order
    sorted_mngrs = sort_pkgmngrs(final_defaults['package']['managers'][OS[0]])
    final_defaults['package']['managers'][OS[0]] = sorted_mngrs

    return final_defaults



#!/usr/bin/env python3.10
"""
       Name: onbaordme.sudo_setup
DESCRIPTION: setup pam module for sudo and add user to sudo group
     AUTHOR: Jesse Hitch
    LICENSE: GNU AFFERO GENERAL PUBLIC LICENSE Version 3
"""

import logging as log
from os import geteuid
from os import system as check_response

# custom libs
from .console_logging import print_header, print_sub_header
from .subproc import subproc


def setup_sudo():
    """
    make sure we're root on mac and kick off setting up sudo with touchid
    Returns True
    """
    print_header("🔒 Setting up sudo")

    # check if running as root
    if geteuid() != 0:
        subproc(["sudo onboardme -s sudo_setup"], spinner=False)
        print_sub_header("🧑‍💻 sudo using TouchId is enabled.")
    else:
        enable_sudo_with_touchid()
    return True


def enable_sudo_with_touchid():
    """
    We look for this line in /etc/pam.d/sudo:
        auth       sufficient     pam_tid.so
    If not found, we add it.
    return True
    """
    pam_file = "/etc/pam.d/sudo"
    if check_response(f'grep "pam_tid.so" {pam_file}') != 0:
        log.info(f"TouchID not found in {pam_file}. Attempting to add it.")

        # read in the file and modify the second line to have pam_tid.so
        new_contents = []
        with open(pam_file, 'r') as file_contents:
            for index, line in enumerate(file_contents.readlines()):
                new_contents.append(line)
                if index == 1:
                    touchid = "auth       sufficient     pam_tid.so\n"
                    new_contents.append(touchid)

        # write back the altered file
        with open(pam_file, 'w') as new_file_contents:
            for line in new_contents:
                new_file_contents.write(line)
    return True



"""
This just handles keeping your home directory as a live git directory
"""
from itertools import zip_longest
from os import chdir, path
from pathlib import Path
# this is for rich text, to pretty print things
from rich import box
from rich.table import Table
# custom libs
from .constants import HOME_DIR, PWD, XDG_CONFIG_DIR
from .console_logging import print_panel, print_msg
from .subproc import subproc


XDG_CONFIG_GIT_PATH = path.join(XDG_CONFIG_DIR, 'git/config')


def setup_dot_files(OS='Linux', overwrite=False,
                    git_url="https://github.com/jessebot/dot_files.git",
                    branch="main") -> None:
    """
    note on how we're doing things, seperate dot files repo:
    https://probablerobot.net/2021/05/keeping-'live'-dotfiles-in-a-git-repo/
    """
    git_dir = path.join(HOME_DIR, '.git_dot_files')
    # create ~/.git_dot_files if it does not exist
    Path(git_dir).mkdir(exist_ok=True)
    chdir(git_dir)
    opts = {'quiet': True, 'cwd': git_dir}

    # global: use main as default branch, always push up new remote branch
    git_raw = "https://raw.githubusercontent.com/jessebot/dot_files/main"

    cmds = [f'curl {git_raw}/.config/git/config -o {XDG_CONFIG_GIT_PATH}',
            f'git --git-dir={git_dir} --work-tree={HOME_DIR} init',
            'git config status.showUntrackedFiles no']
    subproc(cmds, spinner=False, **opts)

    # this one needs to be allowed to fail because it might already exist
    cmds = [f"git remote add origin {git_url}"]
    subproc(cmds, error_ok=True, spinner=False, **opts)

    if overwrite:
        # WARN: this command will overwrite local files with remote files
        reset_cmd = f"git reset --hard origin/{branch}"
    else:
        reset_cmd = f"git reset origin/{branch}"
        git_action = "[b]differ[/b] from"

    # fetch the latest changes, then reset to main, w/o overwriting anything
    subproc(['git fetch', reset_cmd], spinner=False, **opts)

    # get the latest remote modified and deleted files, if there are any
    git_files = subproc([f'git ls-files -m -d {HOME_DIR}'], **opts)

    if overwrite or not git_files:
        # if all the files are updated, just print them all as confirmation :)
        git_cmd = f"git ls-tree --full-tree -r --name-only origin/{branch}"
        git_files = subproc([git_cmd], spinner=False, **opts)
        git_action = "are up to date with"

    print_git_file_table(git_files, git_action, branch, git_url)
    chdir(PWD)

    if not overwrite and 'differ' in git_action:
        # we only print this msg if we got the file exists error
        msg = ("To [warn]:warning: overwrite[/warn] the existing dot files in "
               f"{HOME_DIR}/ with the file(s) listed in the above table, run:"
               "\n[green]onboardme [warn]--overwrite[/warn]")
        print_msg(msg)


def print_git_file_table(remote_git_files=[], file_verb="", branch="",
                         git_url="") -> None:
    """
    Takes a list of files and pretty prints them in a nice table in 2 columns:
        remote_git_files - [], list of files to print in 2 columns
        file_verb        - "", what is their relation to the origin/{branch}
        branch           - "", git branch we're looking at
        git_url          - "", url of git repo we're working with
    """
    emote = "[header]ʕ ･ᴥ･ʔ[/header][sky_blue1]"
    if "differ" in file_verb:
        emote = "[yellow]:warning: ʕ ⚈ᴥ⚈ʔ"
    # table to print the results of all the files
    table = Table(expand=True,
                  box=box.MINIMAL_DOUBLE_HEAD,
                  row_styles=["", "dim"],
                  border_style="dim",
                  show_header=False)
    table.add_column(" ", style="green")

    # remove all trailing space and then create a list of file paths
    # we make this a set to remove the duplicates
    files = set(remote_git_files.rstrip().replace("../..",
                                                  HOME_DIR).split('\n'))
    # then we make this a list so that zip_longest doesn't complain about sets
    f_list = list(files)
    if len(files) < 2:
        table.add_row(f_list[0])
    else:
        table.add_column(" ", style="green")

        # find midpoint of the list. if it's a float (e.g. 20.5) convert to int
        mid = int(len(files) / 2)
        # iterate over both halfs of list till the end of the longest list,
        for (f1, f2) in zip_longest(f_list[0:mid],
                                    f_list[mid:],
                                    fillvalue=" "):
            if f1 != " " and f2 != " ":
                table.add_row(f1, f2)
            else:
                table.add_row(f2, " ")

    git_repo = "[/grn]/[grn]".join(git_url.split('/')[-2:]).replace(".git", "")
    msg = (f"{emote} The following file(s) {file_verb} [grn]"
           f"origin[/grn]/[grn]{branch}[/grn] in [grn]{git_repo}")
    print_panel(table, msg, "left")



# this is for rich text, to pretty print things
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.theme import Theme


CONSOLE = Console(theme=Theme({"warn": "bold yellow",
                               "grn": "medium_spring_green",
                               "ohno": "magenta",
                               "header": "cornflower_blue"}))


def print_manual_steps():
    """
    Just prints out the final steps to be done manually, til we automate them
    """
    # table to print the results of all the files
    table = Table(expand=True, box=None,
                  title=" ",
                  row_styles=["", "dim"],
                  border_style="dim",
                  header_style="cornflower_blue",
                  title_style="light_steel_blue")
    table.add_column("Don't forget these (currently) manual tasks",
                     justify="center")

    table.add_row(" ")
    table.add_row("Import RSS feeds config into FluentReader")
    table.add_row("Import subscriptions into FreeTube")
    table.add_row("⌨️  Set CAPSLOCK to control")
    table.add_row("Install cronjobs you need from ~/.cron")
    table.add_row("Load your BASH config: [green]source .bashrc[/]")
    table.add_row("Reboot, as [turquoise2]docker[/] demands it")
    table.add_row(" ")
    table.add_row("If you need any help, check the docs:")
    table.add_row("[cyan][link=https://jessebot.github.io/onboardme]"
                  "jessebot.github.io/onboardme[/link]")
    table.add_row(" ")

    print_panel(table, '[green]♥ ˖⁺‧Success‧⁺˖ ♥')
    return True


def print_panel(content='', title_txt='', title_alignment='center',
                border_style="light_steel_blue"):
    """
    prints text in a box with a light_steel_blue1 border and title_txt
    """
    print('')
    panel = Panel(content, title=title_txt, title_align=title_alignment,
                  border_style=border_style)
    CONSOLE.print(panel)
    return


def print_header(title='', line_style='royal_blue1'):
    """
    prints text centered in a line that spans the terminal
    """
    print('')
    CONSOLE.rule('[cornflower_blue]' + title, style=line_style)
    return


def print_sub_header(title='', style='italic light_steel_blue', alignment='center'):
    """
    prints text centered in a line that spans the terminal
    """
    print('')
    CONSOLE.print(title, style=style, justify='center')
    return


def print_msg(text='', alignment='center', style='dim italic'):
    """
    prints text centered in the width of the terminal
    """
    CONSOLE.print(text, justify=alignment, style=style)
    return



"""
NAME:    Onboardme.ide_setup
DESC:    install vim, neovim, and fonts
AUTHORS:  Jesse Hitch, Max Roby
LICENSE: GNU AFFERO GENERAL PUBLIC LICENSE
"""

import logging as log
from git import Repo, RemoteProgress
from os import path
from pathlib import Path

# custom libs
from .constants import HOME_DIR, OS
from .console_logging import print_header, print_sub_header, print_msg
from .subproc import subproc


def font_setup() -> None:
    """
    On Linux:
      Clones nerd-fonts repo and does a sparse checkout on only mononoki font.
      Also removes 70-no-bitmaps.conf and links 70-yes-bitmaps.conf

      Then runs install.sh from nerd-fonts repo
    """
    if 'Linux' in OS:
        print_header('📝 [i]font[/i] installations')
        # not sure if needed anymore
        # mkdir -p ~/.local/share/fonts

        fonts_dir = f'{HOME_DIR}/repos/nerd-fonts'

        # do a shallow clone of the repo
        if not path.exists(fonts_dir):
            log.info('Nerdfonts require some setup on Linux...')
            bitmap_conf = '/etc/fonts/conf.d/70-no-bitmaps.conf'
            log.info(f'Going to remove {bitmap_conf} and link a yes map...')
            # we do all of this with subprocess because I want the sudo prompt
            if path.exists(bitmap_conf):
                subproc([f'sudo rm {bitmap_conf}'], quiet=True, spinner=False)

            cmd = ('sudo ln -s /etc/fonts/conf.avail/70-yes-bitmaps.conf '
                   '/etc/fonts/conf.d/70-yes-bitmaps.conf')
            subproc([cmd], error_ok=True, quiet=True, spinner=False)

            print_msg('[i]Downloading installer and font sets... ')

            Path(fonts_dir).mkdir(parents=True, exist_ok=True)
            fonts_repo = 'https://github.com/ryanoasis/nerd-fonts.git'

            class CloneProgress(RemoteProgress):
                def update(self, op_code, cur_count, max_count=None,
                           message=''):
                    if message:
                        log.info(message)

            Repo.clone_from(fonts_repo, fonts_dir, progress=CloneProgress(),
                            multi_options=['--sparse', '--filter=blob:none'])
            cmds = ["git sparse-checkout add patched-fonts/Mononoki"]
            subproc(cmds, spinner=True, cwd=fonts_dir)
        else:
            subproc(["git pull"], spinner=True, cwd=fonts_dir)

        subproc(['./install.sh Mononoki'], quiet=True,
                cwd=fonts_dir)

        print_msg('[i][dim]The fonts should be installed, however, you have ' +
                  'to set your terminal font to the new font. I rebooted too.')


def neovim_setup() -> None:
    """
    Installs all plugins and syncs them if needed.
    Runs this command that works via the cli:
        nvim --headless "+Lazy! sync" +qa
    """
    print_header('[green][i]NeoVim[/i][/green] plugins installation '
                 '[dim]and[/dim] upgrades via [green]lazy.nvim[/green]')

    subproc(['nvim --headless "+Lazy! sync" +qa'])

    print_sub_header('NeoVim Plugins installed.')
